require 'roadforest/source-rigor/graph-store'
require 'roadforest/graph/etagging'
require 'roadforest/graph/access-manager'
require 'roadforest/graph/graph-focus'

module RoadForest
  module Interface
    class ProcessingSequenceError < StandardError
    end

    class Application
      def initialize(route_name, params, path_provider, services)
        @route_name = route_name
        @params = params
        @path_provider = path_provider
        @services = services
        @data = nil
        @response_values = {}
      end
      attr_reader :route_name, :params, :services, :data, :path_provider
      attr_reader :response_values

      #@!group Utility methods

      def path_for(route_name = nil, params = nil)
        path_provider.path_for(route_name, params || self.params)
      end

      def url_for(route_name, params = nil)
        ::RDF::URI.new(Addressable::URI.parse(canonical_host.to_s).join(path_for(route_name, params)))
      end

      def pattern_for(route_name, vals = nil, extra = nil)
        path_provider.pattern_for(route_name, vals, extra)
      end

      def interface_for(route_name = nil, params = nil)
        path_provider.interface_for(route_name, params || self.params)
      end

      def canonical_uri
        url_for(route_name, params)
      end

      def my_path
        path_for(route_name, params)
      end

      def my_url
        canonical_uri.to_s
      end

      #@!endgroup

      #@!group Authorization

      def required_grants(method)
        services.authz.build_grants do |grants|
          grants.add(:admin)
        end
      end

      def authorization(request)
        method = request.method
        required = required_grants(method)
        if required.empty?
          :public
        else
          services.authz.authorization(request, required)
        end
      end

      def authentication_challenge
        services.authz.challenge(:realm => "Roadforest")
      end

      #@!endgroup

      def canonical_host
        services.canonical_host
      end

      def reset #XXX remove?
      end

      #group Resource interface

      def error_data(status)
        nil
      end

      def exists?
        !data.nil?
      end

      def etag
        nil
      end

      def last_modified
        nil
      end

      def response_location
        @response_values.fetch(:location) do
          raise ProcessingSequenceError, "Location not available until request processed"
        end
      end

      def response_location=(location)
        @response_values[:location] = location
      end

      def response_data
        @response_values.fetch(:data) do
          raise ProcessingSequenceError, "Location not available until request processed"
        end
      end

      def response_data=(data)
        @response_values[:data] = data
      end

      def processed
        [:location, :data].each do |key|
          unless @response_values.has_key?(key)
            @response_values[key] = nil
          end
        end
      end

      def expires
        nil
      end

      def update_payload
        nil
      end

      def create_payload
        nil
      end

      def update(data)
        raise NotImplementedError
      end

      def add_child(results)
        raise NotImplementedError
      end

      def retrieve
        raise NotImplementedError
      end
      alias retreive retrieve

      def delete
        false
      end
    end
  end
end
